import abc
import asyncio
import contextlib
from typing import AsyncContextManager, AsyncIterator, Callable, Dict, List, Set, Union

import yarl

from .base import ClosableResponse, Request, Response
from .deadline import Deadline
from .delays_provider import linear_delays
from .priority import Priority
from .request_sender import RequestSender
from .response_classifier import ResponseClassifier, ResponseVerdict
from .utils import cancel_futures, close, close_futures


class RequestStrategy(abc.ABC):
    __slots__ = ()

    @abc.abstractmethod
    def request(
        self,
        request: Request,
        deadline: Deadline,
        priority: Priority = Priority.NORMAL,
    ) -> AsyncContextManager[Response]:
        ...


class MethodBasedStrategy(RequestStrategy):
    __slots__ = ("_strategy_by_method",)

    def __init__(self, strategy_by_method: Dict[str, RequestStrategy]):
        self._strategy_by_method = strategy_by_method

    def request(
        self,
        request: Request,
        deadline: Deadline,
        priority: Priority = Priority.NORMAL,
    ) -> AsyncContextManager[Response]:
        return self._strategy_by_method[request.method].request(request, deadline, priority)


class RequestStrategiesFactory:
    __slots__ = (
        "_request_sender",
        "_endpoint",
        "_response_classifier",
    )

    def __init__(
        self,
        request_sender: RequestSender,
        endpoint: Union[str, yarl.URL],
        response_classifier: ResponseClassifier,
    ):
        self._request_sender = request_sender
        self._endpoint = yarl.URL(endpoint) if isinstance(endpoint, str) else endpoint
        self._response_classifier = response_classifier

    def sequential(
        self, *, attempts_count: int = 3, delays_provider: Callable[[int], float] = linear_delays()
    ) -> RequestStrategy:
        return SequentialRequestStrategy(
            request_sender=self._request_sender,
            endpoint=self._endpoint,
            response_classifier=self._response_classifier,
            attempts_count=attempts_count,
            delays_provider=delays_provider,
        )

    def parallel(
        self, *, attempts_count: int = 3, delays_provider: Callable[[int], float] = linear_delays()
    ) -> RequestStrategy:
        return ParallelRequestStrategy(
            request_sender=self._request_sender,
            endpoint=self._endpoint,
            response_classifier=self._response_classifier,
            attempts_count=attempts_count,
            delays_provider=delays_provider,
        )


class SequentialRequestStrategy(RequestStrategy):
    __slots__ = (
        "_endpoint",
        "_request_sender",
        "_response_classifier",
        "_attempts_count",
        "_delays_provider",
    )

    def __init__(
        self,
        *,
        request_sender: RequestSender,
        endpoint: yarl.URL,
        response_classifier: ResponseClassifier,
        attempts_count: int,
        delays_provider: Callable[[int], float],
    ):
        if attempts_count < 1:
            raise RuntimeError("Attempts count should be >= 1")

        self._endpoint = endpoint
        self._request_sender = request_sender
        self._response_classifier = response_classifier
        self._delays_provider = delays_provider
        self._attempts_count = attempts_count

    @contextlib.asynccontextmanager
    async def request(
        self, request: Request, deadline: Deadline, priority: Priority = Priority.NORMAL
    ) -> AsyncIterator[Response]:
        responses: List[ClosableResponse] = []
        try:
            for attempt in range(self._attempts_count):
                response = await self._request_sender.send(self._endpoint, request, deadline, priority)
                responses.append(response)
                if self._response_classifier.classify(response) == ResponseVerdict.ACCEPT:
                    break
                if attempt + 1 == self._attempts_count:
                    break
                retry_delay = self._delays_provider(attempt + 1)
                if deadline.timeout < retry_delay:
                    break
                await asyncio.sleep(retry_delay)
            yield responses[-1]
        finally:
            await asyncio.shield(close(responses))


class ParallelRequestStrategy(RequestStrategy):
    __slots__ = (
        "_endpoint",
        "_request_sender",
        "_response_classifier",
        "_attempts_count",
        "_delays_provider",
    )

    def __init__(
        self,
        *,
        request_sender: RequestSender,
        endpoint: yarl.URL,
        response_classifier: ResponseClassifier,
        attempts_count: int,
        delays_provider: Callable[[int], float],
    ):
        if attempts_count < 1:
            raise RuntimeError("Attempts count should be >= 1")

        self._endpoint = endpoint
        self._request_sender = request_sender
        self._response_classifier = response_classifier
        self._delays_provider = delays_provider
        self._attempts_count = attempts_count

    async def _schedule_request(
        self, attempt: int, request: Request, deadline: Deadline, priority: Priority = Priority.NORMAL
    ) -> ClosableResponse:
        if attempt > 0:
            await asyncio.sleep(min(self._delays_provider(attempt), deadline.timeout))
        return await self._request_sender.send(self._endpoint, request, deadline, priority)

    @contextlib.asynccontextmanager
    async def request(
        self, request: Request, deadline: Deadline, priority: Priority = Priority.NORMAL
    ) -> AsyncIterator[Response]:
        completed_tasks: Set[asyncio.Future[ClosableResponse]] = set()
        pending_tasks: Set[asyncio.Future[ClosableResponse]] = set()
        for attempt in range(0, self._attempts_count):
            pending_tasks.add(asyncio.create_task(self._schedule_request(attempt, request, deadline, priority)))

        accepted_responses: List[ClosableResponse] = []
        not_accepted_responses: List[ClosableResponse] = []
        try:
            try:
                while pending_tasks:
                    completed_tasks, pending_tasks = await asyncio.wait(
                        pending_tasks, return_when=asyncio.FIRST_COMPLETED
                    )
                    while completed_tasks:
                        completed_task = completed_tasks.pop()
                        response = await completed_task
                        if self._response_classifier.classify(response) == ResponseVerdict.ACCEPT:
                            accepted_responses.append(response)
                        else:
                            not_accepted_responses.append(response)
                    if accepted_responses:
                        break
            finally:
                await asyncio.shield(cancel_futures(pending_tasks))

            yield accepted_responses[0] if accepted_responses else not_accepted_responses[0]
        finally:
            await asyncio.shield(close_futures(pending_tasks | completed_tasks))
            await asyncio.shield(close(accepted_responses + not_accepted_responses))



# flake8: noqa
__version__ = "0.0.24"

from .base import ClosableResponse, EmptyResponse, Header, Method, Request, Response
from .client import Client, setup
from .context import get_context, set_context
from .deadline import Deadline
from .delays_provider import constant_delays, linear_delays
from .metrics import MetricsProvider, NoopMetricsProvider
from .priority import Priority
from .request_sender import RequestSender
from .requests import delete, get, post, post_json, put, put_json
from .response_classifier import DefaultResponseClassifier, ResponseClassifier, ResponseVerdict
from .strategy import MethodBasedStrategy, RequestStrategiesFactory, RequestStrategy
from .transport import Transport

try:
    import aiohttp

    from .aiohttp import AioHttpTransport, aiohttp_middleware_factory
except ImportError:
    pass

try:
    import prometheus_client

    from .prometheus import PROMETHEUS_METRICS_PROVIDER, PrometheusMetricsProvider
except ImportError:
    pass



import asyncio
import json
import logging
from typing import Any, Awaitable, Callable, Optional

import aiohttp
import aiohttp.web_middlewares
import aiohttp.web_request
import aiohttp.web_response
import async_timeout
import multidict
import yarl

from .base import ClosableResponse, EmptyResponse, Header, Request
from .context import set_context
from .deadline import Deadline
from .priority import Priority
from .transport import Transport
from .utils import substitute_path_parameters

logger = logging.getLogger(__package__)


class AioHttpTransport(Transport):
    __slots__ = ("_client_session", "_network_errors_code", "_buffer_payload")

    def __init__(
        self,
        client_session: aiohttp.ClientSession,
        *,
        network_errors_code: int = 489,
        buffer_payload: bool = True,
    ):
        self._client_session = client_session
        self._network_errors_code = network_errors_code
        self._buffer_payload = buffer_payload

    async def send(self, endpoint: yarl.URL, request: Request, timeout: float) -> ClosableResponse:
        if not endpoint.is_absolute():
            raise RuntimeError("Base url should be absolute")

        method = request.method
        url = endpoint.join(substitute_path_parameters(request.url, request.path_parameters))
        headers = request.headers
        body = request.body

        try:
            logger.debug(
                "Sending request %s %s with timeout %s",
                method,
                url,
                timeout,
                extra={
                    "request_method": method,
                    "request_url": url,
                    "request_timeout": timeout,
                },
            )
            response = await self._client_session.request(
                method,
                url,
                headers=headers,
                data=body,
                timeout=timeout,
            )
            if self._buffer_payload:
                await response.read()  # force response to buffer its body
            return _AioHttpResponse(response)
        except aiohttp.ClientError:
            logger.warning(
                "Request %s %s has failed",
                method,
                url,
                exc_info=True,
                extra={
                    "request_method": method,
                    "request_url": url,
                },
            )
            return EmptyResponse(status=self._network_errors_code)
        except asyncio.TimeoutError:
            logger.warning(
                "Request %s %s has timed out after %s",
                method,
                url,
                timeout,
                extra={
                    "request_method": method,
                    "request_url": url,
                    "request_timeout": timeout,
                },
            )
            return EmptyResponse(status=408)


class _AioHttpResponse(ClosableResponse):
    __slots__ = ("_response",)

    def __init__(self, response: aiohttp.ClientResponse):
        self._response = response

    async def close(self) -> None:
        self._response.close()

    @property
    def status(self) -> int:
        return self._response.status

    @property
    def headers(self) -> multidict.CIMultiDictProxy[str]:
        return self._response.headers

    async def json(
        self,
        *,
        encoding: Optional[str] = None,
        loads: Callable[[str], Any] = json.loads,
        content_type: Optional[str] = "application/json",
    ) -> Any:
        return await self._response.json(encoding=encoding, loads=loads, content_type=content_type)

    async def read(self) -> bytes:
        return await self._response.read()

    async def text(self, encoding: Optional[str] = None) -> str:
        return await self._response.text(encoding=encoding)


_HANDLER = Callable[[aiohttp.web_request.Request], Awaitable[aiohttp.web_response.StreamResponse]]
_MIDDLEWARE = Callable[[aiohttp.web_request.Request, _HANDLER], Awaitable[aiohttp.web_response.StreamResponse]]


def aiohttp_middleware_factory(
    *,
    default_timeout: float = 20,
    default_priority: Priority = Priority.NORMAL,
    low_timeout_threshold: float = 0.005,
) -> _MIDDLEWARE:
    @aiohttp.web_middlewares.middleware
    async def middleware(
        request: aiohttp.web_request.Request, handler: _HANDLER
    ) -> aiohttp.web_response.StreamResponse:
        deadline = Deadline.try_parse(request.headers.get(Header.X_REQUEST_DEADLINE_AT)) or Deadline.from_timeout(
            default_timeout
        )
        priority = Priority.try_parse(request.headers.get(Header.X_REQUEST_PRIORITY)) or default_priority

        if deadline.expired or deadline.timeout <= low_timeout_threshold:
            return aiohttp.web_response.Response(status=408)

        with set_context(deadline=deadline, priority=priority):
            try:
                async with async_timeout.timeout(timeout=deadline.timeout):
                    return await handler(request)
            except asyncio.TimeoutError:
                return aiohttp.web_response.Response(status=408)

    return middleware



import random
from typing import Callable

DELAY_PROVIDER = Callable[[int], float]


def constant_delays(*, delay: float = 0) -> DELAY_PROVIDER:
    return lambda _: delay


def linear_delays(
    *, min_delay_seconds: float = 0, delay_multiplier: float = 0.05, jitter: float = 0.2
) -> DELAY_PROVIDER:
    def _linear(attempt: int) -> float:
        delay = min_delay_seconds + attempt * delay_multiplier
        jitter_amount = delay * random.random() * jitter
        if random.random() < 0.5:
            jitter_amount = -jitter_amount
        return delay + jitter_amount

    return _linear



import asyncio
import contextlib
from typing import Any, Collection, Dict, Mapping, Optional, Protocol, TypeVar, Union

import multidict
import yarl

EMPTY_HEADERS = multidict.CIMultiDictProxy[str](multidict.CIMultiDict[str]())


def get_headers_to_enrich(
    headers: Optional[Union[Mapping[Union[str, multidict.istr], str], multidict.CIMultiDictProxy[str]]]
) -> multidict.CIMultiDict[str]:
    return multidict.CIMultiDict[str](headers) if headers is not None else multidict.CIMultiDict[str]()


class Closable(Protocol):
    async def close(self) -> None:
        ...


TClosable = TypeVar("TClosable", bound=Closable)


async def _close(item: TClosable) -> None:
    with contextlib.suppress(Exception):
        await item.close()


async def close_futures(items: Collection[asyncio.Future[TClosable]]) -> None:
    for item in items:
        if item.cancelled():
            continue
        try:
            await _close(await item)
        except asyncio.CancelledError:
            if not item.cancelled():
                raise


async def close(items: Collection[TClosable]) -> None:
    for item in items:
        await _close(item)


async def cancel_futures(futures: Collection[asyncio.Future[TClosable]]) -> None:
    for future in futures:
        if future.done():
            continue
        future.cancel()


def substitute_path_parameters(url: yarl.URL, parameters: Optional[Mapping[str, str]] = None) -> yarl.URL:
    if not parameters:
        return url

    path = url.path
    for name, value in parameters.items():
        path = path.replace(f"{{{name}}}", value)

    build_parameters: Dict[str, Any] = dict(
        scheme=url.scheme,
        user=url.user,
        password=url.password,
        host=url.host,
        port=url.port,
        path=path,
        query=url.query,
        fragment=url.fragment,
    )

    return yarl.URL.build(**{k: v for k, v in build_parameters.items() if v is not None})



import abc
from typing import Dict


class MetricsProvider(abc.ABC):
    __slots__ = ()

    @abc.abstractmethod
    def increment_counter(self, name: str, tags: Dict[str, str], value: float = 1) -> None:
        pass

    @abc.abstractmethod
    def observe_value(self, name: str, tags: Dict[str, str], value: float) -> None:
        pass


class NoopMetricsProvider(MetricsProvider):
    __slots__ = ()

    def increment_counter(self, name: str, tags: Dict[str, str], value: float = 1) -> None:
        pass

    def observe_value(self, name: str, tags: Dict[str, str], value: float) -> None:
        pass



import abc
import json
from typing import Any, Callable, Mapping, Optional, Union

import multidict
import yarl

from .utils import EMPTY_HEADERS, get_headers_to_enrich


class Method:
    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    DELETE = "DELETE"


class Header:
    CONTENT_TYPE = multidict.istr("Content-Type")
    X_REQUEST_DEADLINE_AT = multidict.istr("X-Request-Deadline-At")
    X_REQUEST_PRIORITY = multidict.istr("X-Request-Priority")


class Request:
    __slots__ = (
        "method",
        "url",
        "path_parameters",
        "headers",
        "body",
    )

    def __init__(
        self,
        method: str,
        url: yarl.URL,
        path_parameters: Optional[Mapping[str, str]] = None,
        headers: Optional[multidict.CIMultiDictProxy[str]] = None,
        body: Optional[bytes] = None,
    ):
        if url.is_absolute():
            raise RuntimeError("Request url should be relative")

        self.method = method
        self.path_parameters = path_parameters
        self.url = url
        self.headers = headers
        self.body = body

    def update_headers(
        self, headers: Union[Mapping[Union[str, multidict.istr], str], multidict.CIMultiDictProxy[str]]
    ) -> "Request":
        updated_headers = get_headers_to_enrich(self.headers)
        updated_headers.update(headers)
        return Request(
            method=self.method,
            url=self.url,
            path_parameters=self.path_parameters,
            headers=multidict.CIMultiDictProxy[str](updated_headers),
            body=self.body,
        )


class Response(abc.ABC):
    __slots__ = ()

    @property
    @abc.abstractmethod
    def status(self) -> int:
        ...

    @property
    @abc.abstractmethod
    def headers(self) -> multidict.CIMultiDictProxy[str]:
        ...

    @abc.abstractmethod
    async def json(
        self,
        *,
        encoding: Optional[str] = None,
        loads: Callable[[str], Any] = json.loads,
        content_type: Optional[str] = "application/json",
    ) -> Any:
        ...

    @abc.abstractmethod
    async def read(self) -> bytes:
        ...

    @abc.abstractmethod
    async def text(self, encoding: Optional[str] = None) -> str:
        ...

    def is_informational(self) -> bool:
        return 100 <= self.status < 200

    def is_successful(self) -> bool:
        return 200 <= self.status < 300

    def is_redirection(self) -> bool:
        return 300 <= self.status < 400

    def is_client_error(self) -> bool:
        return 400 <= self.status < 500

    def is_server_error(self) -> bool:
        return 500 <= self.status < 600


class ClosableResponse(Response):
    __slots__ = ()

    @abc.abstractmethod
    async def close(self) -> None:
        ...


class EmptyResponse(ClosableResponse):
    __slots__ = ("_status",)

    def __init__(self, *, status: int):
        self._status = status

    @property
    def status(self) -> int:
        return self._status

    @property
    def headers(self) -> multidict.CIMultiDictProxy[str]:
        return EMPTY_HEADERS

    async def json(
        self,
        *,
        encoding: Optional[str] = None,
        loads: Optional[Callable[[str], Any]] = None,
        content_type: Optional[str] = "application/json",
    ) -> Any:
        return None

    async def read(self) -> bytes:
        return bytes()

    async def text(self, encoding: Optional[str] = None) -> str:
        return ""

    async def close(self) -> None:
        pass



import asyncio
import time

import yarl

from .base import ClosableResponse, EmptyResponse, Header, Request
from .deadline import Deadline
from .metrics import MetricsProvider
from .priority import Priority
from .transport import Transport


class RequestSender:
    __slots__ = ("_transport", "_metrics_provider", "_low_timeout_threshold", "_emit_system_headers")

    def __init__(
        self,
        *,
        transport: Transport,
        metrics_provider: MetricsProvider,
        emit_system_headers: bool = True,
        low_timeout_threshold: float = 0.05,
    ):
        self._transport = transport
        self._metrics_provider = metrics_provider
        self._emit_system_headers = emit_system_headers
        self._low_timeout_threshold = low_timeout_threshold

    async def send(
        self, endpoint: yarl.URL, request: Request, deadline: Deadline, priority: Priority
    ) -> ClosableResponse:
        if deadline.expired or deadline.timeout < self._low_timeout_threshold:
            return EmptyResponse(status=408)

        if self._emit_system_headers:
            request = request.update_headers(
                {
                    Header.X_REQUEST_DEADLINE_AT: str(deadline),
                    Header.X_REQUEST_PRIORITY: str(priority),
                }
            )

        started_at = time.perf_counter()
        try:
            response = await self._transport.send(endpoint, request, deadline.timeout)
            self._capture_metrics(endpoint, request, response.status, started_at)
            return response
        except asyncio.CancelledError:
            self._capture_metrics(endpoint, request, 499, started_at)
            raise

    def _capture_metrics(self, endpoint: yarl.URL, request: Request, status: int, started_at: float) -> None:
        tags = {
            "request_endpoint": endpoint.human_repr(),
            "request_method": request.method,
            "request_path": request.url.path,
            "response_status": str(status),
        }
        elapsed = max(0.0, time.perf_counter() - started_at)
        self._metrics_provider.increment_counter("aio_request_status", tags)
        self._metrics_provider.observe_value("aio_request_latency", tags, elapsed)



import datetime
from typing import Optional


class Deadline:
    @staticmethod
    def from_timeout(seconds: float) -> "Deadline":
        return Deadline(datetime.datetime.utcnow() + datetime.timedelta(seconds=seconds))

    @staticmethod
    def try_parse(value: Optional[str]) -> Optional["Deadline"]:
        if value is None:
            return None
        try:
            return Deadline(datetime.datetime.fromisoformat(value))
        except ValueError:
            return None

    __slots__ = ("deadline_at",)

    deadline_at: datetime.datetime

    def __init__(self, deadline_at: datetime.datetime):
        if deadline_at.tzinfo is not None:
            raise RuntimeError("Deadline should not be zone aware")

        self.deadline_at = deadline_at

    @property
    def timeout(self) -> float:
        return max((self.deadline_at - datetime.datetime.utcnow()).total_seconds(), 0.001)  # 0 is infinite

    @property
    def expired(self) -> bool:
        return (self.deadline_at - datetime.datetime.utcnow()).total_seconds() <= 0

    def __str__(self) -> str:
        return str(self.deadline_at.isoformat())



import abc

import yarl

from .base import ClosableResponse, Request


class Transport(abc.ABC):
    __slots__ = ()

    @abc.abstractmethod
    async def send(self, endpoint: yarl.URL, request: Request, timeout: float) -> ClosableResponse:
        ...



import abc
import enum

from .base import Response


class ResponseVerdict(enum.Enum):
    ACCEPT = 1
    REJECT = 2


class ResponseClassifier(abc.ABC):
    __slots__ = ()

    @abc.abstractmethod
    def classify(self, response: Response) -> ResponseVerdict:
        ...


class DefaultResponseClassifier(ResponseClassifier):
    __slots__ = ("_network_errors_code",)

    def __init__(self, network_errors_code: int = 489):
        self._network_errors_code = network_errors_code

    def classify(self, response: Response) -> ResponseVerdict:
        if response.is_server_error():
            return ResponseVerdict.REJECT
        if response.status == self._network_errors_code:
            return ResponseVerdict.REJECT
        if response.status == 408:
            return ResponseVerdict.REJECT
        return ResponseVerdict.ACCEPT



import re
from pathlib import Path

from setuptools import setup

install_requires = [
    "multidict>=4.5,<7.0",
    "yarl>=1.0,<2.0",
]


def read(*parts):
    return Path(__file__).resolve().parent.joinpath(*parts).read_text().strip()


def read_version():
    regexp = re.compile(r"^__version__\W*=\W*\"([\d.abrc]+)\"")
    for line in read("aio_request", "__init__.py").splitlines():
        match = regexp.match(line)
        if match is not None:
            return match.group(1)
    else:
        raise RuntimeError("Cannot find version in aio_request/__init__.py")


setup(
    name="aio-request",
    version=read_version(),
    description="Various strategies for sending requests",
    platforms=["macOS", "POSIX", "Windows"],
    author="Yury Pliner",
    python_requires=">=3.9",
    project_urls={},
    author_email="yury.pliner@gmail.com",
    license="MIT",
    packages=["aio_request"],
    package_dir={"aio_request": "./aio_request"},
    package_data={"aio_request": ["py.typed"]},
    install_requires=install_requires,
    include_package_data=True,
)



import json
from typing import Any, Callable, Mapping, Optional, Union

import multidict
import yarl

from .base import Header, Method, Request
from .utils import get_headers_to_enrich


def get(
    url: Union[str, yarl.URL],
    *,
    headers: Optional[multidict.CIMultiDictProxy[str]] = None,
    path_parameters: Optional[Mapping[str, str]] = None,
) -> Request:
    return build_request(Method.GET, url, path_parameters=path_parameters, headers=headers)


def post(
    url: Union[str, yarl.URL],
    body: Optional[bytes] = None,
    *,
    headers: Optional[multidict.CIMultiDictProxy[str]] = None,
    path_parameters: Optional[Mapping[str, str]] = None,
) -> Request:
    return build_request(Method.POST, url, path_parameters=path_parameters, headers=headers, body=body)


def put(
    url: Union[str, yarl.URL],
    body: Optional[bytes] = None,
    *,
    headers: Optional[multidict.CIMultiDictProxy[str]] = None,
    path_parameters: Optional[Mapping[str, str]] = None,
) -> Request:
    return build_request(Method.PUT, url, path_parameters=path_parameters, headers=headers, body=body)


def delete(
    url: Union[str, yarl.URL],
    *,
    headers: Optional[multidict.CIMultiDictProxy[str]] = None,
    path_parameters: Optional[Mapping[str, str]] = None,
) -> Request:
    return build_request(Method.DELETE, url, path_parameters=path_parameters, headers=headers)


def post_json(
    url: Union[str, yarl.URL],
    data: Any,
    *,
    path_parameters: Optional[Mapping[str, str]] = None,
    headers: Optional[multidict.CIMultiDictProxy[str]] = None,
    encoding: str = "utf-8",
    dumps: Callable[[str], Any] = json.dumps,
    content_type: str = "application/json",
) -> Request:
    return build_json_request(
        Method.POST,
        url,
        data,
        path_parameters=path_parameters,
        headers=headers,
        encoding=encoding,
        dumps=dumps,
        content_type=content_type,
    )


def put_json(
    url: Union[str, yarl.URL],
    data: Any,
    *,
    path_parameters: Optional[Mapping[str, str]] = None,
    headers: Optional[multidict.CIMultiDictProxy[str]] = None,
    encoding: str = "utf-8",
    dumps: Callable[[str], Any] = json.dumps,
    content_type: str = "application/json",
) -> Request:
    return build_json_request(
        Method.PUT,
        url,
        data,
        path_parameters=path_parameters,
        headers=headers,
        encoding=encoding,
        dumps=dumps,
        content_type=content_type,
    )


def build_json_request(
    method: str,
    url: Union[str, yarl.URL],
    data: Any,
    *,
    path_parameters: Optional[Mapping[str, str]] = None,
    headers: Optional[multidict.CIMultiDictProxy[str]] = None,
    encoding: str = "utf-8",
    dumps: Callable[[Any], str] = json.dumps,
    content_type: str = "application/json",
) -> Request:
    enriched_headers = get_headers_to_enrich(headers)
    enriched_headers.add(Header.CONTENT_TYPE, content_type)

    body = dumps(data).encode(encoding)

    return build_request(
        method=method,
        url=url,
        headers=multidict.CIMultiDictProxy[str](enriched_headers),
        body=body,
        path_parameters=path_parameters,
    )


def build_request(
    method: str,
    url: Union[str, yarl.URL],
    *,
    path_parameters: Optional[Mapping[str, str]] = None,
    headers: Optional[multidict.CIMultiDictProxy[str]] = None,
    body: Optional[bytes] = None,
) -> Request:
    return Request(
        method=method,
        url=yarl.URL(url) if isinstance(url, str) else url,
        headers=headers,
        body=body,
        path_parameters=path_parameters,
    )



from typing import Any, Dict

import prometheus_client

from .metrics import MetricsProvider


class PrometheusMetricsProvider(MetricsProvider):
    __slots__ = ("_metrics", "_registry")

    def __init__(self, registry: prometheus_client.CollectorRegistry) -> None:
        self._metrics: Dict[str, Any] = {}
        self._registry = registry

    def increment_counter(self, name: str, tags: Dict[str, str], value: float = 1) -> None:
        if name not in self._metrics:
            self._metrics[name] = prometheus_client.Counter(name, "", labelnames=tags.keys())
        self._metrics[name].labels(*tags.values()).inc(value)

    def observe_value(self, name: str, tags: Dict[str, str], value: float) -> None:
        if name not in self._metrics:
            self._metrics[name] = prometheus_client.Histogram(name, "", labelnames=tags.keys())
        self._metrics[name].labels(*tags.values()).observe(value)


PROMETHEUS_METRICS_PROVIDER = PrometheusMetricsProvider(prometheus_client.REGISTRY)



import contextlib
import contextvars
import dataclasses
from typing import Iterator, Optional, Union, cast

from .deadline import Deadline
from .priority import Priority

sentinel = object()


@dataclasses.dataclass(frozen=True)
class Context:
    __slots__ = ("deadline", "priority")

    deadline: Optional[Deadline]
    priority: Optional[Priority]

    def set(
        self,
        deadline: Optional[Union[Deadline, object]] = sentinel,
        priority: Optional[Union[Priority, object]] = sentinel,
    ) -> "Context":
        return Context(
            deadline=self.deadline if deadline is sentinel else cast(Optional[Deadline], deadline),
            priority=self.priority if priority is sentinel else cast(Optional[Priority], priority),
        )


context_var = contextvars.ContextVar("aio_request_context", default=Context(None, None))


@contextlib.contextmanager
def set_context(
    *,
    deadline: Optional[Union[Deadline, object]] = sentinel,
    priority: Optional[Union[Deadline, object]] = sentinel,
) -> Iterator[None]:
    reset_token = context_var.set(context_var.get().set(deadline=deadline, priority=priority))
    try:
        yield
    finally:
        context_var.reset(reset_token)


def get_context() -> Context:
    return context_var.get()



from typing import AsyncContextManager, Callable, Optional, Union

import yarl

from .base import Method, Request, Response
from .context import get_context
from .deadline import Deadline
from .delays_provider import linear_delays
from .metrics import MetricsProvider, NoopMetricsProvider
from .priority import Priority
from .request_sender import RequestSender
from .response_classifier import DefaultResponseClassifier, ResponseClassifier
from .strategy import MethodBasedStrategy, RequestStrategiesFactory, RequestStrategy
from .transport import Transport


class Client:
    __slots__ = (
        "_request_strategy",
        "_request_enricher",
        "_default_priority",
        "_default_timeout",
    )

    def __init__(
        self,
        *,
        request_strategy: RequestStrategy,
        default_timeout: float,
        default_priority: Priority,
        request_enricher: Optional[Callable[[Request], Request]],
    ):
        self._default_priority = default_priority
        self._default_timeout = default_timeout
        self._request_enricher = request_enricher
        self._request_strategy = request_strategy

    def request(
        self, request: Request, *, deadline: Optional[Deadline] = None, priority: Optional[Priority] = None
    ) -> AsyncContextManager[Response]:
        if self._request_enricher is not None:
            request = self._request_enricher(request)
        context = get_context()
        return self._request_strategy.request(
            request,
            deadline or context.deadline or Deadline.from_timeout(self._default_timeout),
            self.normalize_priority(priority or self._default_priority, context.priority),
        )

    @staticmethod
    def normalize_priority(priority: "Priority", context_priority: Optional["Priority"]) -> "Priority":
        if context_priority is None:
            return priority

        if priority == Priority.LOW and context_priority == Priority.HIGH:
            return Priority.NORMAL

        if priority == Priority.HIGH and context_priority == Priority.LOW:
            return Priority.NORMAL

        return priority


def setup(
    *,
    transport: Transport,
    endpoint: Union[str, yarl.URL],
    safe_method_attempts_count: int = 3,
    unsafe_method_attempts_count: int = 3,
    safe_method_delays_provider: Callable[[int], float] = linear_delays(delay_multiplier=0.1),
    unsafe_method_delays_provider: Callable[[int], float] = linear_delays(delay_multiplier=0.05),
    response_classifier: Optional[ResponseClassifier] = None,
    default_timeout: float = 20.0,
    default_priority: Priority = Priority.NORMAL,
    low_timeout_threshold: float = 0.005,
    emit_system_headers: bool = True,
    request_enricher: Optional[Callable[[Request], Request]] = None,
    metrics_provider: Optional[MetricsProvider] = None,
) -> Client:
    factory = RequestStrategiesFactory(
        request_sender=RequestSender(
            transport=transport,
            metrics_provider=metrics_provider or NoopMetricsProvider(),
            low_timeout_threshold=low_timeout_threshold,
            emit_system_headers=emit_system_headers,
        ),
        endpoint=endpoint,
        response_classifier=response_classifier or DefaultResponseClassifier(),
    )
    request_strategy = MethodBasedStrategy(
        {
            Method.GET: factory.parallel(
                attempts_count=safe_method_attempts_count, delays_provider=safe_method_delays_provider
            ),
            Method.POST: factory.sequential(
                attempts_count=unsafe_method_attempts_count, delays_provider=unsafe_method_delays_provider
            ),
            Method.PUT: factory.sequential(
                attempts_count=unsafe_method_attempts_count, delays_provider=unsafe_method_delays_provider
            ),
            Method.DELETE: factory.sequential(
                attempts_count=unsafe_method_attempts_count, delays_provider=unsafe_method_delays_provider
            ),
        }
    )
    return Client(
        request_strategy=request_strategy,
        default_timeout=default_timeout,
        default_priority=default_priority,
        request_enricher=request_enricher,
    )



import enum
from typing import Optional


class Priority(str, enum.Enum):
    HIGH = "high"
    NORMAL = "normal"
    LOW = "low"

    def __str__(self) -> str:
        return self.value

    @staticmethod
    def try_parse(value: Optional[str]) -> Optional["Priority"]:
        if value is None:
            return None
        try:
            return Priority(value.lower())
        except ValueError:
            return None
